
#include "app_config.h"

#include <ami_stream.h>
#include <ami_sort.h>

#include "union_find.h"
#include <vector>
#include <ext/hash_map>
#include <ext/hash_set>

using namespace std;
using namespace __gnu_cxx;

/* union-find function calling convention:
	 out is an empty stream already created
	 replist is an empty stream already created, sorted by id upon returning
*/

template<class TIME, class ID>
class cmp_uf_time_t {
public:
	inline int compare(const uf_t<TIME,ID> &c1, const uf_t<TIME,ID> &c2) {
		return (c1.time<c2.time ? -1 : (c2.time<c1.time ? 1 : 0));
	}
};

template<class TIME, class ID>
class cmp_uf_time_id_t {
public:
	inline int compare(const uf_t<TIME,ID> &c1, const uf_t<TIME,ID> &c2) {
		return (c1.time<c2.time ? -1 : (c2.time<c1.time ? 1 :
																		(c1.id<c2.id?-1:(c2.id<c1.id?1:0))));
	}
};

template<class TIME, class ID>
class cmp_uf_id_t {
public:
	inline int compare(const uf_t<TIME,ID> &c1, const uf_t<TIME,ID> &c2) {
		return (c1.id<c2.id ? -1 : (c2.id<c1.id ? 1 : 0));
	}
};

template<class TIME, class ID>
class cmp_uf_id_rtime_t {
public:
	inline int compare(const uf_t<TIME,ID> &c1, const uf_t<TIME,ID> &c2) {
		return (c1.id<c2.id ? -1 : (c2.id<c1.id ? 1 :
																(c2.time<c1.time?-1:(c1.time<c2.time?1:0))));
	}
};

template<class ID>
class cmp_rep_rep_t {
public:
	inline int compare(const rep_t<ID> &c1, const rep_t<ID> &c2) {
		return (c1.rep < c2.rep ? -1 : (c2.rep < c1.rep ? 1 : 0));
	}
};

template<class ID>
class cmp_rep_id_t {
public:
	inline int compare(const rep_t<ID> &c1, const rep_t<ID> &c2) {
		return (c1.id < c2.id ? -1 : (c2.id < c1.id ? 1 : 0));
	}
};

template<class ID>
Node<ID> * OnlineUF<ID>::find(Node<ID> *p)
{
	if (p->parent == p) return p;
	p->parent = p->parent->parent;
	return find(p->parent);
}

template<class ID>
ID OnlineUF<ID>::Find(ID x)
{
	typename hash_map<ID, Node<ID>*>::iterator it0;
	it0 = id2node->find(x);
	if (it0 == id2node->end()) return x;
	else return find(it0->second)->rep;
}

template<class ID>
void OnlineUF<ID>::Union(vector<ID>& ids)
{
	vector<typename hash_map<ID, Node<ID>*>::iterator> it;
	vector<Node<ID>*> v;
	Node<ID>* max_v;

	it.resize(ids.size());
	v.resize(ids.size());
	max_v = NULL;
	unsigned int i;

	for (i=0; i<ids.size(); i++) {
		it[i] = id2node->find(ids[i]);
		if (it[i] == id2node->end()) {
			it[i] = id2node->insert
				(pair<ID, Node<ID>*>(ids[i], new Node<ID>(ids[i]))).first;
		}
		v[i] = find(it[i]->second);
		if (max_v == NULL || v[i]->height > max_v->height) 
			max_v = v[i];				
	}
	for (i=0; i<ids.size(); i++) {
		if (v[i] == max_v) continue;
		v[i]->parent = max_v;
		if (largest && max_v->rep < v[i]->rep ||
				!largest && v[i]->rep < max_v->rep)
			max_v->rep = v[i]->rep;
		if (v[i]->height == max_v->height) 
			max_v->height ++;
	}
}

template<class TIME, class ID>
void uf_mm(AMI_STREAM<uf_t<TIME,ID> > *in, AMI_STREAM<ID> *out, 
					 AMI_STREAM<rep_t<ID> > *replist, bool largest) 
{
	in->seek(0);
	uf_t<TIME,ID> *opp, last_op;
	vector<ID> ids;
	rep_t<ID> r;
	bool end_of_stream;
	OnlineUF<ID> *ouf = new OnlineUF<ID>(largest);

	// do the union-find
	in->read_item(&opp);
	last_op = *opp;
	ids.clear();
	ids.push_back(last_op.id);
	while (1) {
		end_of_stream = true;
		while (in->read_item(&opp) != AMI_ERROR_END_OF_STREAM) {
			if (opp->time == last_op.time) {
				ids.push_back(opp->id);
			} else {
				last_op = *opp;
				end_of_stream = false;
				break;
			}
		}

		if (ids.size() > 1) { // union
			ouf->Union(ids);
		} else { // find
			out->write_item(ouf->Find(ids[0]));
		}

		if (end_of_stream) break;
		ids.clear();
		ids.push_back(last_op.id);
	}
	delete in;

	// build the representative list
	typename hash_map<ID, Node<ID>*>::iterator it0;
	if (replist) {
		it0 = ouf->id2node->begin();
		while (it0 != ouf->id2node->end()) {
			r.id = it0->second->id;
			r.rep = ouf->find(it0->second)->rep;
			replist->write_item(r);
			it0++;
		}
	}

 	delete ouf;

	cmp_rep_id_t<ID> cmp_rep_id;
	if (replist)
		AMI_sort(replist, &cmp_rep_id);
}

template<class TIME, class ID>
void uf(AMI_STREAM<uf_t<TIME,ID> > *in, AMI_STREAM<ID> *out,	
				AMI_STREAM<rep_t<ID> > *replist, bool largest) 
{
	uf_t<TIME,ID> *opp, op, last_op;
	unsigned int M = MM_manager.memory_available() / (24 + sizeof(Node<ID>)) / 2;
	//cout << "Memory = " << MM_manager.memory_available() << " M = " << M << endl;
	TPIE_OS_OFFSET n = in->stream_len();

	unsigned int nn = 0; // # elements touched
	hash_set<ID> *ids1 = new hash_set<ID>;
	in->seek(0);
	while (in->read_item(&opp) != AMI_ERROR_END_OF_STREAM) {
		ids1->insert(opp->id);
		if (ids1->size() > M) break;
	}
	nn = ids1->size();
	delete ids1;
	if (nn < M) {
		uf_mm(in, out, replist, largest);
		return;
	}

	rep_t<ID> *rp, r, r1, r2;
	TIME t;
	TPIE_OS_OFFSET i, j;
	bool b, b1, b2;
	vector<ID> ids;
	bool end_of_stream;
	AMI_STREAM<rep_t<ID> > *replist1, *replist2, *reptemp;
	AMI_STREAM<uf_t<TIME,ID> > *second_half, *temp;

	second_half = new AMI_STREAM<uf_t<TIME,ID> >;
	second_half->persist(PERSIST_DELETE);

	// split into two halves
	in->seek(n/2);
	in->read_item(&opp);
	i = 1;
	t = opp->time;
	while (in->read_item(&opp) != AMI_ERROR_END_OF_STREAM) {
		if (opp->time != t) break;
		i++;
	}
	second_half->write_item(*opp);
	while (in->read_item(&opp) != AMI_ERROR_END_OF_STREAM) {
		op = *opp;
		second_half->write_item(op);
	}
	in->truncate(n/2+i);
	
	// recurse in the first half
	replist1 = new AMI_STREAM<rep_t<ID> >;
	replist1->persist(PERSIST_DELETE);
	uf(in, out, replist1, largest);

	// replace elements in the 2nd half
	cmp_uf_id_t<TIME,ID> cmp_uf_id;
	AMI_sort(second_half, &cmp_uf_id);
	second_half->seek(0);
	replist1->seek(0);
	temp = new AMI_STREAM<uf_t<TIME,ID> >;
	temp->persist(PERSIST_DELETE);
	replist1->read_item(&rp);
	r = *rp;
	b = true;
	while (second_half->read_item(&opp) != AMI_ERROR_END_OF_STREAM) {
		op = *opp;
		while (b && r.id < op.id) {
			if (replist1->read_item(&rp) == AMI_ERROR_END_OF_STREAM) 
				b = false;
			else
				r = *rp;
		}
		if (r.id == op.id) {
			op.id = r.rep;
		}
		temp->write_item(op);
	}

	// remove redundant unions
	cmp_uf_time_t<TIME,ID> cmp_uf_time;
	AMI_sort(temp, &cmp_uf_time);
	temp->seek(0);
	second_half->truncate(0);
	second_half->seek(0);
	temp->read_item(&opp);
	last_op = *opp;
	t = opp->time;
	ids.clear();
	ids.push_back(last_op.id);
	while (1) {
		end_of_stream = true;
		while (temp->read_item(&opp) != AMI_ERROR_END_OF_STREAM) {
			if (opp->time == last_op.time) {
				ids.push_back(opp->id);
			}	else {
				last_op = *opp;
				end_of_stream = false;
				break;
			}
		}
		i = 0;
		if (ids.size() > 1) {
			j = 1;
			while (j<ids.size()) {
				if (!(ids[i] == ids[j])) {
					i++;
					ids[i] = ids[j];
				}
				j++;
			}
			if (i>0) ids.resize(i+1);
			else ids.clear();
		}
		op.time = t;
		for (j=0; j<ids.size(); j++) {
			op.id = ids[j];
			second_half->write_item(op);
		}
		if (end_of_stream) break;
		ids.clear();
		ids.push_back(last_op.id);
		t = last_op.time;
	}
	delete temp;

	if (replist) {
		// recurse in the second half
		replist2 = new AMI_STREAM<rep_t<ID> >;
		replist2->persist(PERSIST_DELETE);
		uf(second_half, out, replist2, largest);

		// dereference replist1
		cmp_rep_rep_t<ID> cmp_rep_rep;
		AMI_sort(replist1, &cmp_rep_rep);
		reptemp = new AMI_STREAM<rep_t<ID> >;
		reptemp->persist(PERSIST_DELETE);
		replist1->seek(0);
		replist2->seek(0);
		replist2->read_item(&rp);
		r1 = *rp;
		b = true;
		while (replist1->read_item(&rp) != AMI_ERROR_END_OF_STREAM) {
			r = *rp;
			while (b && r1.id < r.rep) {
				if (replist2->read_item(&rp) == AMI_ERROR_END_OF_STREAM) 
					b = false;
				else
					r1 = *rp;
			}
			if (r.rep == r1.id) {
				r.rep = r1.rep;
			}
			reptemp->write_item(r);
		}
		delete replist1;
		replist1 = reptemp;
		cmp_rep_id_t<ID> cmp_rep_id;
		AMI_sort(replist1, &cmp_rep_id);
	
		// merge replist1 and replist2 into replist
		replist1->seek(0);
		replist2->seek(0);
		b1 = (replist1->read_item(&rp) != AMI_ERROR_END_OF_STREAM);
		if (b1) r1 = *rp;
		b2 = (replist2->read_item(&rp) != AMI_ERROR_END_OF_STREAM);
		if (b2) r2 = *rp;
		while (b1 || b2) {
			if (b1 && (!b2 || r1.id < r2.id)) {
					replist->write_item(r1);
				b1 = (replist1->read_item(&rp) != AMI_ERROR_END_OF_STREAM);
				if (b1) r1 = *rp;
			} else if (b2 && (!b1 || r2.id < r1.id)) {
				replist->write_item(r2);
				b2 = (replist2->read_item(&rp) != AMI_ERROR_END_OF_STREAM);
				if (b2) r2 = *rp;
			} else {
					replist->write_item(r1);
				b1 = (replist1->read_item(&rp) != AMI_ERROR_END_OF_STREAM);
				if (b1) r1 = *rp;
				b2 = (replist2->read_item(&rp) != AMI_ERROR_END_OF_STREAM);
				if (b2) r2 = *rp;			
			}
		}
		delete replist1;
		delete replist2;
	}	else {
		delete replist1;
		// recurse in the second half
		uf(second_half, out, (AMI_STREAM<rep_t<ID> >*)NULL, largest);
	}
}

template<class TIME, class ID>
void union_find(AMI_STREAM<uf_t<TIME,ID> > *in, AMI_STREAM<ID> *out, 
								bool largest) 
{
	uf(in, out, (AMI_STREAM<rep_t<ID> >*)NULL, largest);
}

